package ub.pfg.preyrobot.vision;

import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.Point;
import org.opencv.core.MatOfByte;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;
import org.opencv.video.Video;
import android.util.Log;

public class OpticalFlowTracker extends Tracker {
	private static final String TAG = "OpticalFlowTracker";
	
	private final int BINARY_THRESHOLD = 50; 
	
	// Amarillo filtra mejor los otros colores
	private final Scalar UPPER_THRESHOLD = new Scalar( 65, 360, 360 ); // 100 y 80
	private final Scalar LOWER_THRESHOLD = new Scalar( 55, 120, 160 );
	
	private static final int 	DETECTION = 1,
								TRACKING = 2;
	
	private final int FACTOR = 4;
	
	private Mat hsv, binary, aux, prev, gray;

	private Size reduced_size;
	
	private int state;
	
	private MatOfPoint2f[] 		points; 		// tracked features
	private MatOfByte 			status;
	private MatOfFloat 			error;
	
	private int count;
	
	public void initDetector() {
		super.initDetector();
		
		hsv = new Mat();
		binary = new Mat();
		aux = new Mat();
		prev = new Mat();
		gray = new Mat();
		
		status = new MatOfByte();
	    error = new MatOfFloat();
		points = new MatOfPoint2f[2];
		points[0] = new MatOfPoint2f();
		points[1] = new MatOfPoint2f();
		
		state = DETECTION;
		count = 0;
		
		reduced_size = new Size( WIDTH/FACTOR, HEIGHT/FACTOR );
	}
	
	public Mat processFrame( Mat rgba ) {		
		switch( state ) {
			case DETECTION:
				Imgproc.cvtColor( rgba, aux, Imgproc.COLOR_RGB2HSV_FULL );
				
				Imgproc.resize( aux, hsv, reduced_size );
				
				Core.inRange( hsv, LOWER_THRESHOLD, UPPER_THRESHOLD, hsv );
				
				Core.bitwise_or( hsv, hsv, binary );
				
				computePosition();
				
				if( !lost ) {
					Imgproc.cvtColor( rgba, prev, Imgproc.COLOR_RGB2GRAY );
				}
				
				break;
			
			case TRACKING:
				count++;
				Imgproc.cvtColor( rgba, gray, Imgproc.COLOR_RGB2GRAY );
				flowTracking();
				gray.copyTo( prev );
				break;
		}
		
		if( !lost ) {
			Core.circle( rgba, currentPosition, 10, new Scalar( 255, 0, 0 ), -1, 8, 0 );
		}
		
		computeDirectionAndFactor();
		RI.sendProportionalMovement( direction, prop_factor );
		
		updatePrevPosition();
		
		if( count > 100 ) {
			count = 0;
			state = DETECTION;
		}
		return rgba;
	}
	
	protected void computePosition() {
		int x, y, countX, countY;
		x = 0; y = 0; countX = 0; countY = 0;
		
		for( int i = 0; i < binary.rows(); i++ ) {
			for( int j = 0; j < binary.cols(); j++ ) {
				double a = binary.get( i, j )[0];
				
				if( a > BINARY_THRESHOLD ) {
					x += j;
					y += i;
					countX++;
					countY++;
				}
			}
		}
		if( countX != 0 && countY != 0) {
			currentPosition.x = (x/countX)*FACTOR;
			currentPosition.y = (y/countY)*FACTOR;
			
			points[0].fromArray( currentPosition );
			state = TRACKING;
			lost = false;
		} else {
			currentPosition.x = 0;
			currentPosition.y = 0;
			lost = true;
		}
	}
	
	public void flowTracking() {
		Video.calcOpticalFlowPyrLK( prev, gray, points[0], points[1], status, error );
		Point[] p = points[1].toArray();
		byte[] a = status.toArray();
		if( a[0] > 0 ) { // comprobacion del status de la feature (1 encontrada)
			currentPosition.x = p[0].x;
			currentPosition.y = p[0].y;
			points[0] = points[1];
		} else {
			state = DETECTION;
		}
	}
	
	public void stop() {
		super.stop();
		
		if( hsv != null) {
			hsv.release();
			hsv = null;
		}
	}

}
